<?php

namespace Drupal\webform_number_limit\Plugin\WebformHandler;

use Drupal\Component\Render\FormattableMarkup;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Render\Element;
use Drupal\webform\Plugin\WebformElement\Number;
use Drupal\webform\Plugin\WebformElementInterface;
use Drupal\webform\Plugin\WebformHandlerBase;
use Drupal\webform\WebformMessageManagerInterface;
use Drupal\webform\WebformSubmissionForm;
use Drupal\webform\WebformSubmissionInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Webform number limit handler.
 *
 * @WebformHandler(
 *   id = "number_limit",
 *   label = @Translation("Number limit"),
 *   category = @Translation("Options"),
 *   description = @Translation("Define number submission limits."),
 *   cardinality = \Drupal\webform\Plugin\WebformHandlerInterface::CARDINALITY_UNLIMITED,
 *   results = \Drupal\webform\Plugin\WebformHandlerInterface::RESULTS_IGNORED,
 *   submission = \Drupal\webform\Plugin\WebformHandlerInterface::SUBMISSION_REQUIRED,
 * )
 */
class NumberLimitWebformHandler extends WebformHandlerBase {

  /**
   * The database object.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected $database;

  /**
   * The webform element plugin manager.
   *
   * @var \Drupal\webform\Plugin\WebformElementManagerInterface
   */
  protected $elementManager;

  /**
   * The webform message manager.
   *
   * @var \Drupal\webform\WebformMessageManagerInterface
   */
  protected $messageManager;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    $instance = parent::create($container, $configuration, $plugin_id, $plugin_definition);
    $instance->database = $container->get('database');
    $instance->elementManager = $container->get('plugin.manager.webform.element');
    $instance->messageManager = $container->get('webform.message_manager');

    return $instance;
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration(): array {
    return [
      'element_key' => '',
      'limit' => NULL,
      'limit_reached_message' => 'The maximum amount of submissions has been reached.',
      'submission_over_limit_message' => "The maximum amount of submissions has been reached. You're limited to @limit, please adjust this number accordingly.",
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function getSummary(): array {
    $settings = $this->getSettings();

    $element = $this->getWebform()->getElement($settings['element_key']);
    if ($element) {
      $webform_element = $this->elementManager->getElementInstance($element);
      $t_args = [
        '@title' => $webform_element->getAdminLabel($element),
        '@type' => $webform_element->getPluginLabel(),
      ];
      $settings['element_key'] = $this->t('@title (@type)', $t_args);
    }
    elseif (empty($settings['element_key'])) {
      $settings['element_key'] = [
        '#type' => 'link',
        '#title' => $this->t('Please add a new number element.'),
        '#url' => $this->getWebform()->toUrl('edit-form'),
      ];
    }
    else {
      $settings['element_key'] = [
        '#markup' => $this->t("'@element_key' is missing.", ['@element_key' => $settings['element_key']]),
        '#prefix' => '<b class="color-error">',
        '#suffix' => '</b>',
      ];
    }

    return [
      '#settings' => $settings,
    ] + parent::getSummary();
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state): array {
    $this->applyFormStateToConfiguration($form_state);

    $form['element_key'] = [
      '#type' => 'select',
      '#title' => $this->t('Element'),
      '#options' => $this->getElements(),
      '#default_value' => $this->configuration['element_key'],
      '#required' => TRUE,
      '#empty_option' => (empty($this->configuration['element_key'])) ? $this->t('- Select -') : NULL,
    ];

    $form['limit'] = [
      '#type' => 'number',
      '#title' => $this->t('Limit'),
      '#description' => $this->t('The number of submissions allowed for this element. The current sum of submission values is @total.', [
        '@total' => $this->getTotal(),
      ]),
      '#min' => 0,
      '#default_value' => $this->configuration['limit'],
    ];

    $form['limit_reached_message'] = [
      '#type' => 'textarea',
      '#title' => $this->t('Limit reached message'),
      '#description' => $this->t('This message will be displayed when the limit is reached.')
        . '<br/><br/>'
        . $this->t('Leave blank to hide this message.'),
      '#default_value' => $this->configuration['limit_reached_message'],
    ];

    $form['submission_over_limit_message'] = [
      '#type' => 'textarea',
      '#title' => $this->t('Submission goes over the limit message'),
      '#description' => $this->t('This message will be displayed when a submission goes over the limit.')
        . '<br/><br/>'
        . $this->t('Leave blank to hide this message.'),
      '#default_value' => $this->configuration['submission_over_limit_message'],
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state): void {
    parent::submitConfigurationForm($form, $form_state);
    $this->applyFormStateToConfiguration($form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function alterForm(array &$form, FormStateInterface $form_state, WebformSubmissionInterface $webform_submission): void {
    $webform = $this->getWebform();

    $formObject = $form_state->getFormObject();
    assert($formObject instanceof WebformSubmissionForm);

    $webformSubmission = $formObject->getEntity();
    assert($webformSubmission instanceof WebformSubmissionInterface);

    $this->messageManager->setWebformSubmission($webformSubmission);
    $this->setWebformSubmission($webform_submission);

    if ($this->getTotal() >= $this->getLimit()) {
      // Make sure any messages are prepended to the form.
      if (!isset($form['webform_message'])) {
        $form['webform_message'] = [
          '#weight' => -100,
        ];
      }

      if ($webform->access('submission_update_any')) {
        $form['webform_message'][] = [
          '#type' => 'webform_message',
          '#message_message' => $this->configuration['limit_reached_message'],
          '#message_type' => 'warning',
          '#weight' => -100,
        ];
        $form['webform_message'][] = [
          '#type' => 'webform_message',
          '#message_message' => $this->messageManager->get(WebformMessageManagerInterface::ADMIN_CLOSED),
          '#message_type' => 'info',
          '#weight' => -100,
        ];
      }
      else {
        $form = [];
        $form['webform_message'][] = [
          '#type' => 'webform_message',
          '#message_message' => $this->configuration['limit_reached_message'],
          '#message_type' => 'warning',
          '#weight' => -100,
        ];
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function alterElement(array &$element, FormStateInterface $form_state, array $context) {
    if (empty($element['#webform_key'])
      || $element['#webform_key'] !== $this->configuration['element_key']) {
      return;
    }

    // Set #max on the number element to make sure
    // this submission won't surpass the limit.
    $remaining = $this->getLimit() - $this->getTotal();
    $element['#max'] = $remaining;

    // Set webform submission for form object.
    /** @var \Drupal\webform\WebformSubmissionForm $form_object */
    $form_object = $form_state->getFormObject();

    /** @var \Drupal\webform\WebformSubmissionInterface $webform_submission */
    $webform_submission = $form_object->getEntity();
    $this->setWebformSubmission($webform_submission);

    // Add post-submit validation to make sure
    // this submission doesn't surpass the limit.
    $element['#element_validate'][] = [$this, 'validateElement'];
  }

  /**
   * Validate webform element limit.
   */
  public function validateElement(&$element, FormStateInterface $form_state, &$complete_form) {
    // Skip if element is not visible.
    if (!Element::isVisibleElement($element)) {
      return;
    }

    $total = $this->getTotal();
    $newTotal = $total + (int) $element['#value'];
    $limit = $this->getLimit();

    if ($newTotal > $limit) {
      $form_state->setError($element, new FormattableMarkup($this->configuration['submission_over_limit_message'], [
        '@total' => $total,
        '@limit' => $limit,
        '@overLimit' => $newTotal - $limit,
        '@remaining' => $limit - $total,
      ]));
    }
  }

  /**
   * Get key/value array of webform number elements.
   *
   * @return array
   *   A key/value array of webform number elements.
   */
  protected function getElements(): array {
    $webform = $this->getWebform();
    $elements = $webform->getElementsInitializedAndFlattened();

    $options = [];
    foreach ($elements as $element) {
      $webform_element = $this->elementManager->getElementInstance($element);

      if ($webform_element instanceof Number) {
        $webform_key = $element['#webform_key'];
        $t_args = [
          '@title' => $webform_element->getAdminLabel($element),
          '@type' => $webform_element->getPluginLabel(),
        ];
        $options[$webform_key] = $this->t('@title (@type)', $t_args);
      }
    }

    // Exclude elements being used by other number limit handlers.
    $handlers = $webform->getHandlers();
    foreach ($handlers as $handler) {
      if ($handler instanceof NumberLimitWebformHandler
        && $handler->getHandlerId() !== $this->getHandlerId()) {
        $element_key = $handler->getSetting('element_key');
        unset($options[$element_key]);
      }
    }

    return $options;
  }

  /**
   * Get the configured limit.
   */
  protected function getLimit(): int {
    return (int) $this->configuration['limit'];
  }

  /**
   * Get selected webform element label.
   */
  protected function getElementLabel(): string {
    $element = $this->getElement();
    $webform_element = $this->getWebformElement();
    return $webform_element->getLabel($element);
  }

  /**
   * Get the selected element.
   */
  protected function getElement(): array {
    return $this->getWebform()->getElement($this->configuration['element_key']);
  }

  /**
   * Get the selected webform element plugin instance.
   */
  protected function getWebformElement(): ?WebformElementInterface {
    $element = $this->getElement();
    return ($element) ? $this->elementManager->getElementInstance($element) : NULL;
  }

  /**
   * Get base query for number limit totals.
   *
   * @return bool|\Drupal\Core\Database\Query\SelectInterface
   *   The base query for number limit totals.
   */
  protected function getTotal(): int {
    $webform = $this->getWebform();

    /** @var \Drupal\Core\Database\StatementInterface $result */
    $query = $this->database->select('webform_submission', 's');
    $query->join('webform_submission_data', 'sd', 's.sid = sd.sid');
    $query->addExpression('SUM(sd.value)', 'sum');
    $query->condition('sd.name', $this->configuration['element_key']);
    $query->condition('sd.webform_id', $webform->id());

    return (int) $query->execute()->fetchField();
  }

}
