<?php

namespace Drupal\webform_number_limit\Plugin\WebformHandler;

use Drupal\Core\Form\FormStateInterface;
use Drupal\webform\Plugin\WebformElement\Number;
use Drupal\webform\WebformSubmissionInterface;
use Drupal\webform_options_limit\Plugin\WebformHandler\OptionsLimitWebformHandler;

/**
 * Webform handler that limits the number of submissions for options/checkboxes.
 *
 * @WebformHandler(
 *   id = "options_number_limit",
 *   label = @Translation("Options/Checkbox with number limit"),
 *   category = @Translation("Options"),
 *   description = @Translation("Define options/checkbox with number submission limits."),
 *   cardinality = \Drupal\webform\Plugin\WebformHandlerInterface::CARDINALITY_UNLIMITED,
 *   results = \Drupal\webform\Plugin\WebformHandlerInterface::RESULTS_IGNORED,
 *   submission = \Drupal\webform\Plugin\WebformHandlerInterface::SUBMISSION_REQUIRED,
 * )
 */
class OptionsNumberLimitWebformHandler extends OptionsLimitWebformHandler {

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration(): array {
    return [
      'number_element_key' => '',
    ] + parent::defaultConfiguration();
  }

  /**
   * {@inheritdoc}
   */
  public function getSummary(): array {
    $summary = parent::getSummary();
    $settings = $this->getSettings();

    $element = $this->getWebform()->getElement($settings['number_element_key']);
    if ($element) {
      $webform_element = $this->elementManager->getElementInstance($element);
      $t_args = [
        '@title' => $webform_element->getAdminLabel($element),
        '@type' => $webform_element->getPluginLabel(),
      ];
      $summary['#settings']['number_element_key'] = $this->t('@title (@type)', $t_args);
    }
    elseif (empty($settings['number_element_key'])) {
      $summary['#settings']['number_element_key'] = [
        '#type' => 'link',
        '#title' => $this->t('Please add a new number element.'),
        '#url' => $this->getWebform()->toUrl('edit-form'),
      ];
    }
    else {
      $summary['#settings']['number_element_key'] = [
        '#markup' => $this->t("'@element_key' is missing.", ['@element_key' => $settings['number_element_key']]),
        '#prefix' => '<b class="color-error">',
        '#suffix' => '</b>',
      ];
    }

    return $summary;
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state): array {
    // Number element settings.
    $form['number_element_settings'] = [
      '#type' => 'details',
      '#title' => $this->t('Number element settings'),
      '#open' => TRUE,
      '#weight' => -1,
    ];

    $form['number_element_settings']['number_element_key'] = [
      '#type' => 'select',
      '#title' => $this->t('Number element'),
      '#options' => $this->getNumberElements(),
      '#default_value' => $this->configuration['number_element_key'],
      '#required' => TRUE,
      '#empty_option' => (empty($this->configuration['number_element_key'])) ? $this->t('- Select -') : NULL,
    ];

    $form = parent::buildConfigurationForm($form, $form_state);
    $form['element_settings']['#weight'] = -2;

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function alterForm(array &$form, FormStateInterface $form_state, WebformSubmissionInterface $webform_submission) {
    parent::alterForm($form, $form_state, $webform_submission);

    $optionsElement = &$form['elements'][$this->configuration['element_key']];
    $numberElement = &$form['elements'][$this->configuration['number_element_key']];

    $option = $form_state->getValue($this->configuration['element_key'])
      ?? $form['elements'][$this->configuration['element_key']]['#default_value']
      ?? NULL;

    if (isset($option)) {
      // Set #max on the number element to make sure this submission won't surpass the limit.
      $max = $this->getOptionsLimits()[$option]['remaining'];
      $numberElement['#max'] = $max;

      // Update the number element value if it's greater than the limit.
      $value = $form_state->getValue($this->configuration['number_element_key']);
      if ($value > $max) {
        $numberElement['#value'] = $max;
      }
    }

    // Add ajax to update the number element.
    $id = 'webform-options-number-limit';
    static::buildAjaxElementWrapper($id, $form['elements']);
    static::buildAjaxElementTrigger($id, $optionsElement);
    static::buildAjaxElementTrigger($id, $numberElement);
    static::buildAjaxElementUpdate($id, $form['elements']);
  }

  /**
   * {@inheritdoc}
   */
  protected function getOptionsTotals(array $values = []) {
    $totals = [];
    $options = $this->getElementOptions();

    foreach ($options as $key => $option) {
      $numberTotal = $this->getNumberElementTotal($this->configuration['element_key'], $key);
      $totals[$key] = $numberTotal;
    }

    return $totals;
  }

  /**
   * Get key/value array of webform number elements.
   *
   * @return array
   *   A key/value array of webform number elements.
   */
  protected function getNumberElements(): array {
    $webform = $this->getWebform();
    $elements = $webform->getElementsInitializedAndFlattened();

    $options = [];
    foreach ($elements as $element) {
      $webform_element = $this->elementManager->getElementInstance($element);

      if ($webform_element instanceof Number) {
        $webform_key = $element['#webform_key'];
        $t_args = [
          '@title' => $webform_element->getAdminLabel($element),
          '@type' => $webform_element->getPluginLabel(),
        ];
        $options[$webform_key] = $this->t('@title (@type)', $t_args);
      }
    }

    // Exclude elements being used by other number limit handlers.
    $handlers = $webform->getHandlers();
    foreach ($handlers as $handler) {
      if ($handler instanceof NumberLimitWebformHandler
        && $handler->getHandlerId() !== $this->getHandlerId()) {
        $element_key = $handler->getSetting('number_element_key');
        unset($options[$element_key]);
      }
    }

    return $options;
  }

  /**
   * Get the total number of submissions for the number element.
   */
  protected function getNumberElementTotal(string $elementName, string $elementOption): int {
    $webform = $this->getWebform();

    /** @var \Drupal\Core\Database\StatementInterface $result */
    $query = $this->database->select('webform_submission', 's');
    $query->join('webform_submission_data', 'sd', 's.sid = sd.sid');
    $query->join('webform_submission_data', 'sda', 'sd.sid = sda.sid AND sda.name = :element_name AND sda.value = :element_option', [
      ':element_name' => $elementName,
      ':element_option' => $elementOption,
    ]);
    $query->addExpression('SUM(sd.value)', 'sum');
    $query->condition('sd.name', $this->configuration['number_element_key']);
    $query->condition('sd.webform_id', $webform->id());

    return (int) $query->execute()->fetchField();
  }

}
